import numpy as np

from radardef.types import ExpDef

mu_exp = ExpDef(
    name="mu_experiment",
    radar_frequency=46.5,
    t_ipp_usec=3120,
    t_samp_usec=6,
    t_rx_start_usec=486,
    t_rx_end_usec=486 + 6 * 85,
    t_tx_start_usec=0,
    t_tx_end_usec=26 * 6,  # (code length * (baud_length/t_samp)) * t_samp
    baud_length_usec=12,
    code=np.array(
        [1, 1, 1, 1, 1, -1, -1, 1, 1, -1, 1, -1, 1],
        dtype=np.float64,
    ),
    rx_channels=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25],
    samples_per_file=266240,
)



from radardef.radar_stations.eiscat.experiments import load_radar_code
from radardef.types import ExpDef

leo_sat_1_0l = ExpDef(
    name="leo_sat_1.0l",
    radar_frequency=930,  # TODO
    t_ipp_usec=20000,
    t_samp_usec=2,
    t_tx_start_usec=82,
    t_tx_end_usec=2002,
    t_rx_start_usec=0,
    t_rx_end_usec=20000,
    baud_length_usec=30,
    code=load_radar_code("leo_bpark"),
    t_cal_on_usec=19900,
    t_cal_off_usec=19997,
    samples_per_file=12800000,
    rx_channels=["uhf"],
    tx_channel="uhf",
)

leo_sat_1_0u = ExpDef(
    name="leo_sat_1.0u",
    radar_frequency=930,  # TODO
    t_ipp_usec=20000,
    t_samp_usec=2,
    t_tx_start_usec=82,
    t_tx_end_usec=2002,
    t_rx_start_usec=0,
    t_rx_end_usec=20000,
    baud_length_usec=30,
    code=load_radar_code("leo_bpark"),
    t_cal_on_usec=19900,
    t_cal_off_usec=19997,
    samples_per_file=12800000,
    rx_channels=["uhf"],
    tx_channel="uhf",
)



import configparser
import importlib.resources
import logging
import pathlib

import numpy as np
from numpy.typing import NDArray

from radardef.radar_stations.eiscat.experiments.constants import ExpGroup, ExpVer
from radardef.types import ExpDef

logger = logging.getLogger(__name__)

EISCAT_EXPERIMENTS: dict[ExpGroup, dict[ExpVer, ExpDef]] = {group: {} for group in ExpGroup}
EXP_FILES: dict[str, pathlib.Path] = {}

# To be compatible with 3.7-8
# as resources.files was introduced in 3.9
if hasattr(importlib.resources, "files"):
    _data_files = importlib.resources.files("radardef.radar_stations.eiscat.experiments")
    for file in _data_files.iterdir():
        if isinstance(file, pathlib.Path):
            if not file.is_file():
                continue
            if file.name.endswith(".py"):
                continue

            EXP_FILES[file.name] = file.resolve()

else:
    _data_content = importlib.resources.contents("radardef.radar_stations.eiscat.experiments")
    for fname in _data_content:
        with importlib.resources.path("radardef.radar_stations.eiscat.experiments", fname) as file:
            if not file.is_file():
                continue
            if file.name.endswith(".py"):
                continue

            EXP_FILES[file.name] = pathlib.Path(str(file)).resolve()


def load_radar_code(xpname: ExpGroup | str) -> NDArray[np.float64]:
    """Load radar code, xpname + _code.txt."""

    code_name = xpname + "_code.txt"
    assert code_name in EXP_FILES, f"radar code '{code_name}' not found in pre-defined configurations"
    code_file = EXP_FILES[code_name]
    try:
        with open(code_file, "r") as fh:
            code = []
            for line in fh:
                code.append([1 if ch == "+" else -1 for ch in line.strip()])
        return np.array(code, dtype=np.float64)
    except Exception as e:
        raise ValueError(f"Couldn't open code file for {xpname}:" + str(e))


try:
    from radardef.radar_stations.eiscat.experiments.leo_bpark import (
        leo_bpark_2_0,
        leo_bpark_2_0u,
        leo_bpark_2_1u,
        leo_bpark_2_2,
        leo_bpark_2_3v,
        leo_bpark_2_4u,
        leo_bpark_2_5u,
    )

    EISCAT_EXPERIMENTS[ExpGroup.leo_bpark][ExpVer.v2_0] = leo_bpark_2_0
    EISCAT_EXPERIMENTS[ExpGroup.leo_bpark][ExpVer.v2_0u] = leo_bpark_2_0u
    EISCAT_EXPERIMENTS[ExpGroup.leo_bpark][ExpVer.v2_1u] = leo_bpark_2_1u
    EISCAT_EXPERIMENTS[ExpGroup.leo_bpark][ExpVer.v2_2] = leo_bpark_2_2
    EISCAT_EXPERIMENTS[ExpGroup.leo_bpark][ExpVer.v2_3v] = leo_bpark_2_3v
    EISCAT_EXPERIMENTS[ExpGroup.leo_bpark][ExpVer.v2_4u] = leo_bpark_2_4u
    EISCAT_EXPERIMENTS[ExpGroup.leo_bpark][ExpVer.v2_5u] = leo_bpark_2_5u
except ImportError:
    logger.warning("Could not import leo_bpark experiments")

try:
    from radardef.radar_stations.eiscat.experiments.leo_mpark import leo_mpark_2_1u

    EISCAT_EXPERIMENTS[ExpGroup.leo_mpark][ExpVer.v2_1u] = leo_mpark_2_1u
except ImportError:
    logger.warning("Could not import leo_mpark experiments")

try:
    from radardef.radar_stations.eiscat.experiments.leo_pwait import (
        leo_pwait_2_3r,
        leo_pwait_2_3u,
        leo_pwait_2_3v,
        leo_pwait_2_4u,
        leo_pwait_2_5u,
    )

    EISCAT_EXPERIMENTS[ExpGroup.leo_mpark][ExpVer.v2_3r] = leo_pwait_2_3r
    EISCAT_EXPERIMENTS[ExpGroup.leo_mpark][ExpVer.v2_3u] = leo_pwait_2_3u
    EISCAT_EXPERIMENTS[ExpGroup.leo_mpark][ExpVer.v2_3v] = leo_pwait_2_3v
    EISCAT_EXPERIMENTS[ExpGroup.leo_mpark][ExpVer.v2_4u] = leo_pwait_2_4u
    EISCAT_EXPERIMENTS[ExpGroup.leo_mpark][ExpVer.v2_5u] = leo_pwait_2_5u

except ImportError:
    logger.warning("Could not import leo_pwait experiments")

try:
    from radardef.radar_stations.eiscat.experiments.leo_sat import leo_sat_1_0l, leo_sat_1_0u

    EISCAT_EXPERIMENTS[ExpGroup.leo_sat][ExpVer.v1_0l] = leo_sat_1_0l
    EISCAT_EXPERIMENTS[ExpGroup.leo_sat][ExpVer.v1_0u] = leo_sat_1_0u

except ImportError:
    logger.warning("Could not import leo_sat experiments")


def get_experiment(group: ExpGroup | str, version: ExpVer | str) -> ExpDef:
    try:
        return EISCAT_EXPERIMENTS[ExpGroup(group)][ExpVer(version)]
    except KeyError:
        breakpoint()
        raise ValueError(f"No experiment with name: {group} and version: {version} available")



"""Template of a Radar station"""

import logging
from pathlib import Path
from typing import Optional

from numpy.typing import NDArray
from pyant import Beam
from pyant.types import Parameters
from spacecoords import celestial, frames, spherical

from radardef.components.converter_template import Converter
from radardef.components.data_loader_template import DataLoader
from radardef.components.validator_template import Validator
from radardef.types import ExpDef, TargetFormat


class RadarStation:
    """
    Definition of a radar station.

    Args:
        station_id: A unique string that identifies a radar station.
        transmitter: If the radar station has a transmitter.
        receiver: If the radar station has a receiver.
        lat: Geographical latitude of radar station in decimal degrees  (North+).
        lon: Geographical longitude of radar station in decimal degrees (East+).
        alt: Geographical altitude above geoid surface of radar station in meter.
        beam: The radar beam model for this station.
        beam_parameters: The parameters for the radar beam model for this station.
        min_elevation (optional): The minimum elevation the radar can measure at.
        noise_temperature (optional): The noise temperature in Kelvins intrinsic to the radar receiver.
        power (optional): The maximum power in Watts the radar transmitter can deliver.
        power_per_element (optional): power_per_element
        frequency (optional): The frequency the radar operates at.
        validator (optional): Validator to validate the correct data format from the station
        converters (optional):  List of converters that can converte from the source format to other formats
        data_loaders (optional):  Data loader object compatible with the converted files

    """

    __logger = logging.getLogger(__name__)

    @property
    def station_id(self) -> str:
        """A unique string that identifies a radar station."""
        return self.__station_id

    @property
    def transmitter(self) -> bool:
        """If the radar station has a transmitter."""
        return self.__transmitter

    @property
    def receiver(self) -> bool:
        """If the radar station has a receiver."""
        return self.__receiver

    @property
    def lat(self) -> float:
        """Geographical latitude of radar station in decimal degrees  (North+)."""
        return self.__lat

    @property
    def lon(self) -> float:
        """Geographical longitude of radar station in decimal degrees (East+)."""
        return self.__lon

    @property
    def alt(self) -> float:
        """Geographical altitude above geoid surface of radar station in meter."""
        return self.__alt

    @property
    def beam(self) -> Beam:
        """The radar beam for this station."""
        return self.__beam

    @property
    def beam_parameters(self) -> Parameters:
        """The parameters for the radar beam model for this station."""
        return self.__beam_parameters

    @property
    def ecef(self) -> NDArray:
        """The ITRS coordinates of the radar station."""
        return self.__ecef

    @property
    def ecef_lat(self) -> float:
        """The latitude of the ITRS coordinates of the radar station."""
        return self.__ecef_lat

    @property
    def ecef_lon(self) -> float:
        """The longitude of the ITRS coordinates of the radar station."""
        return self.__ecef_lon

    @property
    def ecef_alt(self) -> float:
        """The altitude of the ITRS coordinates of the radar station."""
        return self.__ecef_alt

    @property
    def min_elevation(self) -> float | None:
        """The minimum elevation the radar can measure at."""
        return self.__min_elevation

    @property
    def noise_temperature(self) -> float | None:
        """The noise temperature in Kelvins intrinsic to the radar receiver."""
        return self.__noise_temperature

    @property
    def power(self) -> float | None:
        """The maximum power in Watts the radar transmitter can deliver."""
        return self.__power

    @property
    def power_per_element(self) -> float | None:
        """Power per element."""
        return self.__power_per_element

    @property
    def frequency(self) -> float | None:
        """The frequency the radar operates at."""
        return self.__frequency

    @property
    def validator(self) -> Validator | None:
        """Validator to confirm that the"""
        return self.__validator

    def enu(self, ecefs: NDArray) -> NDArray:
        """Converts a set of ECEF states to local ENU coordinates using geocentric zenith."""

        rel_ = ecefs.copy()
        rel_[:3, :] = rel_[:3, :] - self.ecef[:, None]
        rel_[:3, :] = frames.ecef_to_enu(
            self.ecef_lat,
            self.ecef_lon,
            rel_[:3, :],
            degrees=True,
        )
        if ecefs.shape[0] > 3:
            rel_[3:, :] = frames.ecef_to_enu(
                self.ecef_lat,
                self.ecef_lon,
                rel_[3:, :],
                degrees=True,
            )

        return rel_

    def __init__(
        self,
        station_id: str,
        transmitter: bool,
        receiver: bool,
        lat: float,
        lon: float,
        alt: float,
        beam: Beam,
        beam_parameters: Parameters,
        min_elevation: Optional[float] = None,
        noise_temperature: Optional[float] = None,
        power: Optional[float] = None,
        power_per_element: Optional[float] = None,
        frequency: Optional[float] = None,
        converters: Optional[list[Converter]] = None,
        validator: Optional[Validator] = None,
        data_loaders: Optional[list[type[DataLoader]]] = None,
    ) -> None:

        self.__station_id = station_id
        self.__transmitter = transmitter
        self.__receiver = receiver
        self.__lat = lat
        self.__lon = lon
        self.__alt = alt
        self.__beam = beam
        self.__beam_parameters = beam_parameters
        self.__ecef = celestial.geodetic_to_ITRS(lat, lon, alt, degrees=True)
        ecef_lla = spherical.cart_to_sph(self.ecef, degrees=True)
        self.__ecef_lat = ecef_lla[1]
        self.__ecef_lon = 90 - ecef_lla[0]
        self.__ecef_alt = ecef_lla[2]
        self.__min_elevation = min_elevation
        self.__noise_temperature = noise_temperature
        self.__power = power
        self.__power_per_element = power_per_element
        self.__frequency = frequency
        self.__validator = validator
        self.__converters: dict[TargetFormat, Converter] = dict()
        self.__data_loaders: dict[TargetFormat, type[DataLoader]] = dict()

        if converters is not None:
            for converter in converters:
                self.add_converter(converter)

        if data_loaders is not None:
            for data_loader in data_loaders:
                self.add_data_loader(data_loader)

    def get_converters(self) -> list[Converter]:
        """Get all converters connected to this radar"""
        return list(self.__converters.values())

    def get_data_loaders(self) -> list[type[DataLoader]]:
        """Get all data loaders connected to this radar"""
        return list(self.__data_loaders.values())

    def add_converter(self, converter: Converter) -> None:
        """Add a converter to this radar"""
        self.__converters[converter.target_format] = converter

    def add_data_loader(self, data_loader: type[DataLoader]) -> None:
        """Add a data loader to this radar"""
        self.__data_loaders[data_loader.converted_format] = data_loader

    def convert(self, path: Path, target_format: TargetFormat, dst: Path) -> list[Path] | None:
        """Convert data from this radar to a specific format"""

        if self.__validator is not None:
            if not self.__validator.validate(path):
                self.__logger.error("Source format not supported")
                return None
        else:
            self.__logger.error("No source validator available, will try to convert anyways")

        try:
            return self.__converters[target_format].convert(path, dst)
        except KeyError:
            self.__logger.error("Not possible to convert to the wanted format, no converter available")

        return None

    def load_data(
        self, path: Path, converted_format: Optional[TargetFormat] = None, experiment: Optional[ExpDef] = None
    ) -> DataLoader | None:
        """Load converted data from this radar"""
        if converted_format is None:
            for data_loader in self.__data_loaders.values():
                if data_loader.validate(path):
                    loader = data_loader(experiment)
                    loader.load(path)
                    return loader
            self.__logger.error("Source format not supported")
        else:
            try:
                if self.__data_loaders[converted_format].validate(path):
                    loader = self.__data_loaders[converted_format](experiment)
                    loader.load(path)
                    return loader
                else:
                    self.__logger.error("Source format not supported")
            except KeyError:
                self.__logger.error("Not possible to load the wanted format, no loader available")

        return None



import logging
import re
from pathlib import Path
from typing import Optional

import h5py
import numpy as np
import numpy.typing as npt

from radardef.components import DataLoader
from radardef.radar_stations.eiscat.experiments import get_experiment
from radardef.radar_stations.eiscat.utils.drf_utils import ts_from_str
from radardef.radar_stations.eiscat.validators import HDF5
from radardef.types import (
    BoundParams,
    ExpDef,
    Pointing,
    TargetFormat,
)


class HDF5Loader(DataLoader):
    """
    HDF5 layout:
    ```
        DATA <Actual data>
          ├─EndTime
          ├─IntegrationTime
          ├─L1
          ├─L2
          └─ParBlock
                └─ParBlock
        MetaData <Data specs>
          ├─EndTime
          ├─IntegrationTime
          ├─L1
          ├─L2
          └─ParBlock
                └─ParBlockLayout
        PortalDBR <Portal specs>
          ├─AccumulatedSeconds
          ├─DataStream
          ├─Documentation
          ├─ExperimentName
          ├─InfoId
          ├─RequestID
          └─ResourceID
    ```

    """

    __logger = logging.getLogger(__name__)

    converted_format = TargetFormat.HDF5
    validator = HDF5()

    # Data section
    DATA = "Data"
    DATA_LEVEL = "L1"
    REAL_IND = 0
    IMAG_IND = 1
    PARBLOCK = "ParBlock"
    PARBLOCK_ELEVATION = 8
    PARBLOCK_AZIMUTH = 9
    PARBLOCK_FREQUENCY = 54  # Not stated in docs
    ENDTIME = "EndTime"
    INTEGRATIONTIME = "IntegrationTime"
    # PortalDBReference section
    PORTALDBREFERENCE = "PortalDBReference"
    DATASTREAM = "DataStream"
    EXPERIMENTNAME = "ExperimentName"

    @property
    def epoch_bounds(self) -> BoundParams:
        """Data epoch bounds"""
        return self.__epoch_bounds

    @property
    def channels(self) -> list[int] | list[str]:
        """All available channels"""
        return self.experiment.rx_channels

    def __init__(self, exp: Optional[ExpDef] = None) -> None:
        super().__init__(exp)
        self.__epoch_bounds = BoundParams()

    def load(self, path: Path | str) -> None:
        """
        Loads a path to the dataloader, extracting metadata and other important specifications

        Args:
            path: path to data file
        """
        self.__path = Path(path).resolve()
        if not self._experiment:
            file = self._open_hdf5_file(self.__path)
            name = file[self.PORTALDBREFERENCE][self.EXPERIMENTNAME][()][0].decode()
            file.close

            expname, expvers, owner = self._expinfo_split(name)
            self.__experiment = get_experiment(group=expname, version=expvers)

        self.__dumps, self.__samples_per_dump = self._get_data_size(self.__path)
        self.__epoch_bounds = self._extract_bounds(self.__path)
        self.__pointing = self._extract_pointing(self.__path)

    def bounds(self, channel: str | int) -> tuple[int, int]:
        """Sample bounds of the specific channel

        Args:
            channel: channel
        Returns:
            tuple of start and end sample
        Raises:
            Exception: channel is missing
        """
        if not isinstance(channel, str):
            channel = str(channel)

        if self._is_channel_present(channel):
            return (0, self.__dumps * self.__samples_per_dump)
        else:
            raise Exception(f"channel {channel} is missing in {dir}")

    def read(
        self,
        channel: str | int,
        start_sample: Optional[int] = None,
        vector_length: Optional[int] = None,
    ) -> npt.NDArray[np.complex128]:
        """

        HDF5 data format:
        ```
            [dump, [real, imag], samples]
        ```

        Args:
            channel: channel to read data from
            start_sample (optional): sample to start reading from, if not given bounds start will be used.
            vector_length (optional): Amount of samples to read from start_sample,
                                      if not given one batch will be read.

        Returns:
            Complex data of given channel
        """
        if not isinstance(channel, str):
            channel = str(channel)

        if channel not in self.channels:
            raise Exception(f"channel {channel} missing in {dir}")

        if start_sample is None:
            start_sample = 0
        if vector_length is None:
            vector_length = self.__samples_per_dump

        dump_index = start_sample // self.__samples_per_dump
        sample_index = start_sample % self.__samples_per_dump
        windows = -((sample_index + vector_length) // -self.__samples_per_dump)

        file = self._open_hdf5_file(self.__path)

        # Concatenate the dump windows
        raw_data = file[self.DATA][self.DATA_LEVEL][dump_index : dump_index + windows].reshape(2, -1)

        file.close()

        # extract samples
        data = np.empty(vector_length, dtype=complex)
        data.real = raw_data[self.REAL_IND, sample_index : sample_index + vector_length].flatten()
        data.imag = raw_data[self.IMAG_IND, sample_index : sample_index + vector_length].flatten()

        return data

    # TODO: change to carthesian coordinates

    def pointing(self, sample: int) -> Pointing:
        """Pointing data, data describing the radar pointing direction in spherical coordinates"""

        block_id = sample // self.__samples_per_dump
        return Pointing(azimuth=self.__pointing[block_id, 0], elevation=self.__pointing[block_id, 1])

    def _extract_pointing(self, path: Path) -> npt.NDArray:
        """Extract pointing data from the parameter block"""

        file = self._open_hdf5_file(path)
        data = np.zeros((file[self.DATA][self.PARBLOCK][self.PARBLOCK].shape[0], 2))
        for i in range(data.shape[0]):
            data[i, 0] = file[self.DATA][self.PARBLOCK][self.PARBLOCK][i][self.PARBLOCK_AZIMUTH]
            data[i, 1] = file[self.DATA][self.PARBLOCK][self.PARBLOCK][i][self.PARBLOCK_ELEVATION]
        file.close()
        return data

    def _extract_bounds(self, path: Path) -> BoundParams:
        """
        Extract meta data from HDF5 file and experiment config files

        """
        file = self._open_hdf5_file(path)

        if (
            self.experiment.radar_frequency
            != file[self.DATA]["ParBlock"]["ParBlock"][0][self.PARBLOCK_FREQUENCY]
        ):
            raise ValueError("Radar frequency does not match with frequency in measurement file")
        if self.experiment.rx_channels[0] != file[self.PORTALDBREFERENCE][self.DATASTREAM][0].decode():
            raise ValueError("Rx channel does not match with channel in measurement file")

        start_time_sec = (
            ts_from_str(file[self.DATA][self.ENDTIME][0].decode()) - file[self.DATA][self.INTEGRATIONTIME][0]
        )
        end_time_sec = ts_from_str(file[self.DATA][self.ENDTIME][-1].decode())
        bounds = BoundParams(ts_start_usec=int(start_time_sec * 1e6), ts_end_usec=int(end_time_sec * 1e6))
        file.close()
        return bounds

    def _get_data_size(self, path: Path) -> tuple[int, int]:
        """Extract amount of dumps and sample per dump from hdf5 file"""
        file = self._open_hdf5_file(path)

        dumps, _, sample_per_dump = file[self.DATA][self.DATA_LEVEL].shape

        return dumps, sample_per_dump

    def _open_hdf5_file(self, path: Path) -> h5py.File:
        """Open hdf5 file and return reader"""

        try:
            h5file = h5py.File(str(path), "r")
        except FileNotFoundError:
            self.__logger.exception(f"Could not open file: {path}. File does not exist.")
            raise
        except OSError:
            self.__logger.exception(f"File {path} was not a h5 file, and was probably in binary format.")
            raise
        except UnicodeDecodeError:
            self.__logger.exception(f"File {path} was not a h5 file.")
            raise

        return h5file

    def _expinfo_split(self, xpinf: str) -> tuple[str, ...]:
        """
        Move from hard coded constants to loading config based on exp name/version
        'leo_bpark_2.1u_NO' -> (leo_bpark', '2.1u', 'NO')
        """
        try:
            match = re.match(r"(\w+)_(\d+(?:\.\d+)?[a-z]*)_(\w+)", xpinf)
            if match is not None:
                return match.groups()
            else:
                return "", "", ""
        except Exception as e:
            raise ValueError(f"d_ExpInfo: {xpinf} not understood: {e}")

    def _get_channel_from_file_name(self, file_name: str) -> str:
        """Extract channel from file name
        'leo_mpark_2.1u_EI@uhf_2024' -> 'uhf'
        """
        match = re.findall(r"(?<=@)[^_]+(?=_)", file_name)
        return match[0]

    def _is_channel_present(self, chnl: str | int) -> bool:
        """is channel present in the data"""
        return chnl in self.experiment.rx_channels



"""General types"""

from dataclasses import dataclass, field
from typing import NamedTuple, Optional

import numpy as np
import numpy.typing as npt
import scipy.constants


class BoundParams(NamedTuple):
    """
    Time since epoch (standard unix) bounds
    """

    ts_start_usec: float | int = 0
    ts_end_usec: float | int = 0


class Pointing(NamedTuple):
    """
    Pointing
    """

    azimuth: float
    elevation: float


@dataclass(frozen=True)
class ExpDef:
    """
    Experiment defintion
    """

    name: str
    radar_frequency: float
    t_ipp_usec: int
    t_samp_usec: int
    t_rx_start_usec: int
    t_rx_end_usec: int
    t_tx_start_usec: int
    t_tx_end_usec: int
    baud_length_usec: int
    samples_per_file: int
    code: npt.NDArray[np.float64]
    rx_channels: list[str] | list[int]
    tx_channel: Optional[str | int] = None
    t_cal_on_usec: Optional[float] = None
    t_cal_off_usec: Optional[float] = None
    ipp_samps: int = field(init=False)
    sample_rate: float = field(init=False)
    wavelength: float = field(init=False)

    def __post_init__(self) -> None:
        object.__setattr__(self, "sample_rate", 1 / (self.t_samp_usec * 1e-6))
        object.__setattr__(self, "wavelength", scipy.constants.c / (self.radar_frequency * 1e6))
        object.__setattr__(self, "ipp_samps", int(self.t_ipp_usec / self.t_samp_usec))



from enum import auto

# Python StrEnum has default lowercase for auto() but is only available from py 3.11
try:
    from enum import StrEnum
except ImportError:
    from strenum import (  # type: ignore[assignment, no-redef, unused-ignore, import-not-found]
        LowercaseStrEnum as StrEnum,  # type: ignore[import-not-found,no-redef, unused-ignore]
    )


class ExpGroup(StrEnum):
    leo_bpark = auto()
    leo_mpark = auto()
    leo_pwait = auto()
    leo_sat = auto()


class ExpVer(StrEnum):
    v1_0l = "1.0l"
    v1_0u = "1.0u"
    v2_0 = "2.0"
    v2_0u = "2.0u"
    v2_1u = "2.1u"
    v2_2 = "2.2"
    v2_3u = "2.3u"
    v2_3v = "2.3v"
    v2_3r = "2.3r"
    v2_4u = "2.4u"
    v2_5u = "2.5u"



"""
Class and freestanding functions to load metadata and data from a file converted from Eiscat mat.bz2
to DRF format. The data loader is based on the DataLoader template.
"""

import configparser
from pathlib import Path
from typing import Optional

import digital_rf
import numpy as np
import numpy.typing as npt

from radardef.components.data_loader_template import DataLoader
from radardef.radar_stations.eiscat.experiments import get_experiment
from radardef.radar_stations.eiscat.validators import DRF
from radardef.types import (
    Boundparam,
    BoundParams,
    ExpDef,
    Expparam,
    Metaparam,
    Pointing,
    TargetFormat,
)


class DrfLoader(DataLoader):
    """Simplifies the way to load DRF files converted from eiscat"""

    converted_format = TargetFormat.DRF
    validator = DRF()

    @property
    def epoch_bounds(self) -> BoundParams:
        """Get data epoch bounds"""
        return self.__epoch_bounds

    @property
    def channels(self) -> list[int] | list[str]:
        """All available channels"""
        return self.__channel_reader.get_channels()

    def __init__(self, exp: Optional[ExpDef] = None) -> None:
        super().__init__(exp)

    def load(self, path: Path | str) -> None:
        """Loads a path to the dataloader, extracting metadata and other important specifications

        Args:
            path: path to data file

        Raises:
            Exception: path is not a directory
            Exception: pointing dir must be a directory
        """

        self.__path = Path(path).resolve()

        if not self.__path.is_dir():
            raise Exception(f"<dir> must be directory path, {self.__path}")

        if not self._experiment:
            meta_file = configparser.ConfigParser()
            meta_file.read(self.__path / "metadata.ini")
            self._experiment = get_experiment(
                group=meta_file.get(Metaparam.EXPERIMENT, Expparam.NAME),
                version=meta_file.get(Metaparam.EXPERIMENT, Expparam.VERSION),
            )

        self.__channel_reader = digital_rf.DigitalRFReader(str(self.__path))

        pointing_dir = self.__path / "pointing"
        if not pointing_dir.is_dir():
            raise Exception(f"<dir/pointing> must be directory path, {self.__path}")

        self.__meta_reader = digital_rf.DigitalMetadataReader(str(pointing_dir))
        idx_start, idx_end = self.__meta_reader.get_bounds()
        self._pointing_inds = list(self.__meta_reader.read(idx_start, idx_end).keys())
        self.__epoch_bounds = self._extract_bounds(self.__path)

    def bounds(self, channel: str | int) -> tuple[int, int]:
        """Sample bounds of the specific channel

        Args:
            channel: channel
        Returns:
            tuple of start and end sample
        Raises:
            Exception: channel is missing
        """

        if not isinstance(channel, str):
            channel = str(channel)

        if channel not in self.__channel_reader.get_channels():
            raise Exception(f"channel {channel} missing in {dir}")

        idx_first, idx_last = self.__channel_reader.get_bounds(channel)
        return idx_first, idx_last + 1

    def read(
        self,
        channel: str | int,
        start_sample: Optional[int] = None,
        vector_length: Optional[int] = None,
    ) -> npt.NDArray[np.complex128]:
        """
        Read data from loaded file

        Args:
            channel (optional): Channel to read data from
            start_sample (optional): Start of range to read, if empty all data will be read
            vector_length (optional): Number of samples (counting from start sample) to read,
                                    if empty all data will be read

        Returns:
            Complex data of length vector_length from give channel

        Raises:
            Exception: channel is missing
        """

        if not isinstance(channel, str):
            channel = str(channel)

        if channel not in self.__channel_reader.get_channels():
            raise Exception(f"channel {channel} missing in {dir}")

        if start_sample is None or vector_length is None:
            bound_start, bound_end = self.bounds(channel)
            return self.__channel_reader.read_vector_1d(bound_start, bound_end - 1, channel)
        else:
            return self.__channel_reader.read_vector_1d(start_sample, vector_length, channel)

    def pointing(self, sample: int) -> Pointing:
        """Pointing data, data describing the radar pointing direction in spherical coordinates"""
        if self.__meta_reader is None:
            return Pointing(azimuth=0, elevation=0)
        else:
            for p_ind in self._pointing_inds:
                if sample < p_ind:
                    data = self.__meta_reader.read(p_ind)[p_ind]
                    return Pointing(data["azimuth"], data["elevation"])

            data = self.__meta_reader.read(self._pointing_inds[0])[self._pointing_inds[0]]
            return Pointing(data["azimuth"], data["elevation"])

    def _extract_bounds(self, path: Path) -> BoundParams:
        meta_file = configparser.ConfigParser()
        meta_file.read(self.__path / "metadata.ini")

        bounds = BoundParams(
            ts_start_usec=meta_file.getfloat(Metaparam.BOUNDS, Boundparam.TS_START_USEC),
            ts_end_usec=meta_file.getfloat(Metaparam.BOUNDS, Boundparam.TS_END_USEC),
        )

        return bounds



"""Eiscat .hdf5 validator"""

from pathlib import Path

from radardef.components.validator_template import Validator
from radardef.types import TargetFormat


class HDF5(Validator):
    """Eiscat hdf5 validator"""

    def __init__(self) -> None:
        super().__init__(TargetFormat.HDF5)

    def validate(self, src: str | Path) -> bool:
        """Validate file is a Eiscat hdf5 file"""
        path = Path(src).resolve()
        return path.name.endswith(".hdf5")



from radardef.radar_stations.eiscat.experiments import load_radar_code
from radardef.types import ExpDef

leo_pwait_2_3r = ExpDef(
    name="leo_pwait_2.3r",
    radar_frequency=930,
    t_ipp_usec=20000,
    t_samp_usec=1,
    t_tx_start_usec=82,
    t_tx_end_usec=2002,
    t_rx_start_usec=0,
    t_rx_end_usec=20000,
    baud_length_usec=30,
    code=load_radar_code("leo_bpark"),
    t_cal_on_usec=19900,
    t_cal_off_usec=19997,
    samples_per_file=12800000,
    rx_channels=["sod"],
    tx_channel="sod",
)

leo_pwait_2_3u = ExpDef(
    name="leo_pwait_2.3u",
    radar_frequency=930,
    t_ipp_usec=20000,
    t_samp_usec=1,
    t_tx_start_usec=82,
    t_tx_end_usec=2002,
    t_rx_start_usec=0,
    t_rx_end_usec=20000,
    baud_length_usec=30,
    code=load_radar_code("leo_bpark"),
    t_cal_on_usec=19900,
    t_cal_off_usec=19997,
    samples_per_file=12800000,
    rx_channels=["uhf"],
    tx_channel="uhf",
)


leo_pwait_2_3v = ExpDef(
    name="leo_pwait_2.3v",
    radar_frequency=224,
    t_ipp_usec=20000,
    t_samp_usec=1,
    t_tx_start_usec=82,
    t_tx_end_usec=2002,
    t_rx_start_usec=0,
    t_rx_end_usec=20000,
    baud_length_usec=30,
    code=load_radar_code("leo_bpark"),
    t_cal_on_usec=19900,
    t_cal_off_usec=19997,
    samples_per_file=12800000,
    rx_channels=["vhf"],
    tx_channel="vhf",
)


leo_pwait_2_4u = ExpDef(
    name="leo_pwait_2.4u",
    radar_frequency=930,
    t_ipp_usec=20000,
    t_samp_usec=1,
    t_tx_start_usec=82,
    t_tx_end_usec=2002,
    t_rx_start_usec=0,
    t_rx_end_usec=20000,
    baud_length_usec=30,
    code=load_radar_code("leo_bpark"),
    t_cal_on_usec=19900,
    t_cal_off_usec=19997,
    samples_per_file=12800000,
    rx_channels=["uhf"],
    tx_channel="uhf",
)

leo_pwait_2_5u = ExpDef(
    name="leo_pwait_2.5u",
    radar_frequency=930,
    t_ipp_usec=20000,
    t_samp_usec=1,
    t_tx_start_usec=82,
    t_tx_end_usec=2002,
    t_rx_start_usec=0,
    t_rx_end_usec=20000,
    baud_length_usec=30,
    code=load_radar_code("leo_bpark"),
    t_cal_on_usec=19900,
    t_cal_off_usec=19997,
    samples_per_file=12800000,
    rx_channels=["uhf"],
    tx_channel="uhf",
)



import types

from radardef.collections import (
    ConverterCollection,
    DataLoaderCollection,
    FormatCollection,
)
from radardef.components import Converter, DataLoader, RadarStation, Validator
from radardef.download import download
from radardef.radar_def import RadarDef
from radardef.radar_stations import *
from radardef.types import Boundparam, BoundParams, ExpDef, SourceFormat, TargetFormat

from .version import __version__
